//! @file Matrix.hh
//! @brief 行列/ベクトル計算クラス(テンプレート版)
//! @date 2020/03/01
//! @author Yuki YOKOKURA
//
// Copyright (C) 2011-2020 Yuki YOKOKURA
// This program is free software;
// you can redistribute it and/or modify it under the terms of the BSD License.
// For details, see the License.txt file.
//
// 以下，コメント。
// ・各々の関数における計算結果はMATLAB/Maximaと比較して合っていることを確認している。
// ・ただし，LU分解などの答えが一意に求まらない関数では，当然，MATLABとは異なる結果を出力する。
// ・古代的なC/C++の記述法と最新の現代的なC++の記述法が混在しているが，そこはご愛嬌ということで。
// ・動的メモリ版に比べてかなり高速の行列演算が可能。
//
// 今後の予定
// ・double型直書きではなくて，テンプレートに書き換える
// ・QR分解，SVD，rank，固有値，固有ベクトルを返す関数を実装する
// ・複素数に対応させる

#ifndef MATRIX
#define MATRIX

#include <string>
#include <cmath>
#include <cassert>
#include <array>

// ARCS組込み用マクロ
#ifdef ARCS_IN
	// ARCSに組み込まれる場合
	#include "ARCSassert.hh"
#else
	// ARCSに組み込まれない場合
	#define arcs_assert(a) (assert(a))
#endif

#define PrintMatSize(a)  (PrintMatSize_Macro((a),#a))	//!< 行列サイズ表示マクロ
#define PrintMatrix(a,b) (PrintMatrix_Macro((a),b,#a))	//!< 行列要素表示マクロ

namespace ARCS {	// ARCS名前空間
	//! @brief 行列/ベクトル計算クラス(テンプレート版)
	//! @tparam	NN	行列の幅
	//! @tparam MM	行列の高さ
	template <unsigned int NN, unsigned int MM>
	class Matrix {
		public:
			//! @brief LU分解の際の並べ替えが偶数回/奇数回発生したことを返すための定義
			enum LUperm {
				ODD,	//!< 奇数
				EVEN	//!< 偶数
			};
			
			//! @brief コンストラクタ
			Matrix()
				: Nindex(0), Mindex(0), Data({0})
			{
				FillAll(0);	// すべての要素を零で初期化
			}
			
			//! @brief コンストラクタ(任意初期値版)
			//! @param[in]	InitValue	行列要素の初期値
			Matrix(const double& InitValue)
				: Nindex(0), Mindex(0), Data({0})
			{
				FillAll(InitValue);	// すべての要素を指定した値で初期化
			}
			
			//! @brief コンストラクタ(初期化リスト版)
			//! @param[in]	InitList	初期化リスト
			Matrix(std::initializer_list<double> InitList)
				: Nindex(0), Mindex(0), Data({0})
			{
				auto ListVal = InitList.begin();	// 初期化リストの最初のポインタ位置
				for(unsigned int i = 0; i < InitList.size(); ++i){
					// 初期化リストを順番に読み込んでいく
					arcs_assert(Nindex < N);		// 行カウンタが行の長さ以内かチェック
					arcs_assert(Mindex < M);		// 列カウンタが列の長さ以内かチェック
					Data[Nindex][Mindex] = (double)ListVal[i];	// 行列の要素を埋める
					Nindex++;			// 行カウンタをインクリメント
					if(Nindex == N){	// 行カウンタが最後まで行き着いたら，
						Nindex = 0;		// 行カウンタを零に戻して，
						Mindex++;		// その代わりに，列カウンタをインクリメント
					}
				}
				Nindex = 0;	// すべての作業が終わったので，
				Mindex = 0;	// 行，列カウンタを零に戻しておく
			}
			
			/*
			//! @brief ムーブコンストラクタ
			Matrix(Matrix&& r)
				: Nindex(0), Mindex(0), Data(std::move(r.Data))
			{
				
			}
			*/
			
			//! @brief コピーコンストラクタ
			Matrix(const Matrix& right)
				: Nindex(0), Mindex(0), Data(right.Data)
			{
				
			}
			
			//! @brief デストラクタ
			~Matrix(){
				
			}
			
			//! @brief 行列代入演算子
			//! @param[in] right 演算子の右側
			//! @return 結果
			auto& operator=(const auto& right){
				static_assert(this->N == right.N, "Matrix Size Error");	// 行列のサイズチェック
				static_assert(this->M == right.M, "Matrix Size Error");	// 行列のサイズチェック
				for(unsigned int i=0;i<N;i++){
					for(unsigned int j=0;j<M;j++) this->Data[i][j] = right.Data[i][j];
				}
				return (*this);
			}

			//! @brief 単項プラス演算子
			//! @return 結果
			auto operator+(void) const{
				Matrix ret;
				for(unsigned int i=0;i<N;i++){
					for(unsigned int j=0;j<M;j++) ret.Data[i][j] = Data[i][j];
				}
				return ret;
			}
			
			//! @brief 単項マイナス演算子
			//! @return 結果
			auto operator-(void) const{
				Matrix ret;
				for(unsigned int i=0;i<N;i++){
					for(unsigned int j=0;j<M;j++) ret.Data[i][j] = -Data[i][j];
				}
				return ret;
			}
			
			//! @brief 行列加算演算子 (行列同士の加算の場合)
			//! @param[in] right 演算子の右側
			//! @return 結果
			auto operator+(const Matrix& right) const{
				static_assert(N == right.N, "Matrix Size Error");	// 行列のサイズチェック
				static_assert(M == right.M, "Matrix Size Error");	// 行列のサイズチェック
				Matrix ret;
				for(unsigned int i=0;i<N;i++){
					for(unsigned int j=0;j<M;j++) ret.Data[i][j] = Data[i][j] + right.Data[i][j];
				}
				return ret;
			}
			
			//! @brief 行列加算演算子 (行列＋スカラーの場合)
			//! @param[in] right 演算子の右側
			//! @return 結果
			auto operator+(const double& right) const{
				Matrix ret;
				for(unsigned int i=0;i<N;i++){
					for(unsigned int j=0;j<M;j++) ret.Data[i][j] = Data[i][j] + right;
				}
				return ret;
			}
			
			//! @brief 行列減算演算子 (行列同士の減算の場合)
			//! @param[in] right 演算子の右側
			//! @return 結果
			auto operator-(const Matrix& right) const{
				static_assert(N == right.N, "Matrix Size Error");	// 行列のサイズチェック
				static_assert(M == right.M, "Matrix Size Error");	// 行列のサイズチェック
				Matrix ret;
				for(unsigned int i=0;i<N;i++){
					for(unsigned int j=0;j<M;j++) ret.Data[i][j] = Data[i][j] - right.Data[i][j];
				}
				return ret;
			}
			
			//! @brief 行列減算演算子 (行列－スカラーの場合)
			//! @param[in] right 演算子の右側
			//! @return 結果
			auto operator-(const double& right) const{
				Matrix ret;
				for(unsigned int i=0;i<N;i++){
					for(unsigned int j=0;j<M;j++) ret.Data[i][j] = Data[i][j] - right;
				}
				return ret;
			}
			
			//! @brief 行列乗算演算子 (行列同士の乗算の場合)
			//! @param[in] right 演算子の右側
			//! @return 結果
			template <unsigned int Nright, unsigned int Mright>
			auto operator*(const Matrix<Nright,Mright>& right) const{
				static_assert(N == right.M, "Matrix Size Error");	// 行列のサイズチェック
				Matrix<right.N,M> ret;
				for(unsigned int k=0;k<right.N;k++){
					for(unsigned int i=0;i<N;i++){
						for(unsigned int j=0;j<M;j++) ret.Data[k][j] += Data[i][j]*right.Data[k][i];
					}
				}
				return ret;
			}
			
			//! @brief 行列乗算演算子 (行列＊スカラーの場合)
			//! @param[in] right 演算子の右側
			//! @return 結果
			auto operator*(const double& right) const{
				Matrix ret;
				for(unsigned int i=0;i<N;i++){
					for(unsigned int j=0;j<M;j++) ret.Data[i][j] = Data[i][j]*right;
				}
				return ret;
			}
			
			//! @brief 行列スカラー除算演算子 (行列／スカラーの場合)
			//! @param[in] right 演算子の右側
			//! @return 結果
			auto operator/(const double& right) const{
				Matrix ret;
				for(unsigned int i=0;i<N;i++){
					for(unsigned int j=0;j<M;j++) ret.Data[i][j] = Data[i][j]/right;
				}
				return ret;
			}

			//! @brief 行列加算代入演算子 (行列＝行列＋行列の場合)
			//! @param[in] right 演算子の右側
			//! @return 結果
			auto& operator+=(const Matrix& right){
				static_assert(this->N == right.N, "Matrix Size Error");	// 行列のサイズチェック
				static_assert(this->M == right.M, "Matrix Size Error");	// 行列のサイズチェック
				for(unsigned int i=0;i<N;i++){
					for(unsigned int j=0;j<M;j++) this->Data[i][j] = this->Data[i][j] + right.Data[i][j];
				}
				return (*this);
			}
			
			//! @brief 行列加算代入演算子 (行列＝行列＋スカラーの場合)
			//! @param[in] right 演算子の右側
			//! @return 結果
			auto& operator+=(const double& right){
				for(unsigned int i=0;i<N;i++){
					for(unsigned int j=0;j<M;j++) this->Data[i][j] = this->Data[i][j] + right;
				}
				return (*this);
			}
			
			//! @brief 行列減算代入演算子 (行列＝行列－行列の場合)
			//! @param[in] right 演算子の右側
			//! @return 結果
			auto& operator-=(const Matrix& right){
				static_assert(this->N == right.N, "Matrix Size Error");	// 行列のサイズチェック
				static_assert(this->M == right.M, "Matrix Size Error");	// 行列のサイズチェック
				for(unsigned int j=0;j<M;j++){
					for(unsigned int i=0;i<N;i++) this->Data[i][j] = this->Data[i][j] - right.Data[i][j];
				}
				return (*this);
			}
			
			//! @brief 行列減算代入演算子 (行列＝行列－スカラーの場合)
			//! @param[in] right 演算子の右側
			//! @return 結果
			auto& operator-=(const double& right){
				for(unsigned int i=0;i<N;i++){
					for(unsigned int j=0;j<M;j++) this->Data[i][j] = this->Data[i][j] - right;
				}
				return (*this);
			}
			
			//! @brief 行列べき乗演算子
			//! @param[in] right 演算子の右側
			//! @return 結果
			auto operator^(const unsigned int& right){
				Matrix A;
				for(unsigned int i=0;i<N;i++){
					for(unsigned int j=0;j<M;j++) A.Data[i][j] = Data[i][j];
				}
				Matrix ret = A;
				for(unsigned int k=1;k<right;k++) ret = ret*A;
				return ret;
			}
			
			//! @brief 行列アダマール乗算演算子 (行列の要素ごとの乗算)
			//! @param[in] right 演算子の右側
			//! @return 結果
			auto operator&(const Matrix& right) const{
				static_assert(N == right.N, "Matrix Size Error");	// 行列のサイズチェック
				static_assert(M == right.M, "Matrix Size Error");	// 行列のサイズチェック
				Matrix ret;
				for(unsigned int i=0;i<N;i++){
					for(unsigned int j=0;j<M;j++) ret.Data[i][j] = Data[i][j]*right.Data[i][j];
				}
				return ret;
			}
			
			//! @brief 行列アダマール除算演算子 (行列の要素ごとの除算)
			//! @param[in] right 演算子の右側
			//! @return 結果
			auto operator%(const Matrix& right) const{
				static_assert(N == right.N, "Matrix Size Error");	// 行列のサイズチェック
				static_assert(M == right.M, "Matrix Size Error");	// 行列のサイズチェック
				Matrix ret;
				for(unsigned int i=0;i<N;i++){
					for(unsigned int j=0;j<M;j++) ret.Data[i][j] = Data[i][j]/right.Data[i][j];
				}
				return ret;
			}
			
			//! @brief 行列添字演算子（縦ベクトルのN番目の要素の値を返す。GetElement(1,M)と同じ意味。ただしサイズチェックは行わない。）
			//! @param[in]	N	要素番号(「1」始まり)
			//! @return	要素の値
			const double& operator[](unsigned int M) const{
				arcs_assert(M != 0);	// 「0」始まり防止チェック
				return Data[0][M-1];
			}
			
			//! @brief 行列添字演算子（縦ベクトルのN番目の要素に値を設定する。GetElement(1,M)と同じ意味。ただしサイズチェックは行わない。）
			//! @param[in]	N	要素番号(「1」始まり)
			//! @return	要素の値
			double& operator[](unsigned int M){
				arcs_assert(M != 0);	// 「0」始まり防止チェック
				return Data[0][M-1];
			}
			
			//! @brief 行列加算演算子 (スカラー＋行列の場合)
			//! @param[in] left		左側のスカラー値
			//! @param[in] right	右側の行列
			friend Matrix operator+(const double& left, const Matrix& right){
				Matrix ret;
				for(unsigned int i=0;i<right.N;i++){
					for(unsigned int j=0;j<right.M;j++) ret.Data[i][j] = left + right.Data[i][j];
				}
				return ret;
			}
			
			//! @brief 行列減算演算子 (スカラー－行列の場合)
			//! @param[in] left		左側のスカラー値
			//! @param[in] right	右側の行列
			friend Matrix operator-(const double& left, const Matrix& right){
				Matrix<NN,MM> ret;
				for(unsigned int i=0;i<right.N;i++){
					for(unsigned int j=0;j<right.M;j++) ret.Data[i][j] = left - right.Data[i][j];
				}
				return ret;
			}
			
			//! @brief 行列乗算演算子 (スカラー＊行列の場合)
			//! @param[in] left		左側のスカラー値
			//! @param[in] right	右側の行列
			friend Matrix operator*(const double& left, const Matrix& right){
				Matrix<NN,MM> ret;
				for(unsigned int i=0;i<right.N;i++){
					for(unsigned int j=0;j<right.M;j++) ret.Data[i][j] = right.Data[i][j]*left;
				}
				return ret;
			}
			
			//! @brief 行列要素に値を設定する関数
			//! @param[in] u1...u2 要素の値
			template<typename T1, typename... T2>		// 可変長引数テンプレート
			void Set(const T1& u1, const T2&... u2){	// 再帰で順番に可変長引数を読み込んでいく
				arcs_assert(Nindex < N);				// 行カウンタが行の長さ以内かチェック
				arcs_assert(Mindex < M);				// 列カウンタが列の長さ以内かチェック
				Data[Nindex][Mindex] = (double)u1;	// 行列の要素を埋める
				Nindex++;			// 行カウンタをインクリメント
				if(Nindex == N){	// 行カウンタが最後まで行き着いたら，
					Nindex = 0;		// 行カウンタを零に戻して，
					Mindex++;		// その代わりに，列カウンタをインクリメント
				}
				Set(u2...);			// 自分自身を呼び出す(再帰)
			}
			void Set(){
				// 再帰の最後に呼ばれる関数
				Nindex = 0;	// すべての作業が終わったので，
				Mindex = 0;	// 行，列カウンタを零に戻しておく
			}
			
			//! @brief 行列要素から値を読み込む関数
			//! @param[in] u1...u2 要素の値
			template<typename T1, typename... T2>		// 可変長引数テンプレート
			void Get(T1& u1, T2&... u2){	// 再帰で順番に可変長引数を読み込んでいく
				arcs_assert(Nindex < N);	// 行カウンタが行の長さ以内かチェック
				arcs_assert(Mindex < M);	// 列カウンタが列の長さ以内かチェック
				u1 = Data[Nindex][Mindex];	// 行列の要素から読み込み
				Nindex++;			// 行カウンタをインクリメント
				if(Nindex == N){	// 行カウンタが最後まで行き着いたら，
					Nindex = 0;		// 行カウンタを零に戻して，
					Mindex++;		// その代わりに，列カウンタをインクリメント
				}
				Get(u2...);			// 自分自身を呼び出す(再帰)
			}
			void Get(){
				// 再帰の最後に呼ばれる関数
				Nindex = 0;	// すべての作業が終わったので，
				Mindex = 0;	// 行，列カウンタを零に戻しておく
			}
			
			//! @brief 1次元std::array配列を縦ベクトルとして読み込む関数
			//! @param[in]	Array	std::array配列(横1×縦MM)
			void LoadArray(const std::array<double, MM>& Array){
				arcs_assert(N == 1);
				arcs_assert(M == MM);
				for(unsigned int j=0;j<M;j++) Data[0][j] = Array[j];
			}
			
			//! @brief 縦ベクトルを1次元std::array配列に書き込む関数
			//! @param[out]	Array	std::array配列(横1×縦MM)
			void StoreArray(std::array<double, MM>& Array){
				arcs_assert(N == 1);
				arcs_assert(M == MM);
				for(unsigned int j=0;j<M;j++) Array[j] = Data[0][j];
			}
			
			//! @brief 2次元std::array配列を読み込む関数
			//! @param[in]	Array	std::array配列(横NN×縦MM)
			void LoadArray(const std::array<std::array<double, NN>, MM>& Array){
				arcs_assert(N == NN);
				arcs_assert(M == MM);
				for(unsigned int i=0;i<N;i++){
					for(unsigned int j=0;j<M;j++) Data[i][j] = Array[j][i];
				}
			}

			//! @brief 指定した要素番号に値を設定する関数 (並び順逆版)
			//! @param[in]	m行目
			//! @param[in]	n列目
			//! @param[in]	valを代入
			void SetElem(unsigned int m, unsigned int n, double val){
				arcs_assert(1<=n && n<=N && 1<=m && m<=M);	// サイズチェック
				Data[n-1][m-1] = val;
			}
			
			//! @brief 指定した要素番号の値を返す関数 (並び順逆版)
			//! @param[in]	m行目
			//! @param[in]	n列目
			//! @return	要素の値
			double GetElem(unsigned int m, unsigned int n) const {
				arcs_assert(1<=n && n<=N && 1<=m && m<=M);	// サイズチェック
				return Data[n-1][m-1];
			}
			
			//! @brief 指定した要素番号に値を設定する関数
			//! @param[in]	n列目
			//! @param[in]	m行目
			//! @param[in]	valを代入
			void SetElement(unsigned int n, unsigned int m, double val){
				arcs_assert(1<=n && n<=N && 1<=m && m<=M);	// サイズチェック
				Data[n-1][m-1] = val;
			}
			
			//! @brief 指定した要素番号の値を返す関数
			//! @param[in]	n列目
			//! @param[in]	m行目
			//! @return	要素の値
			double GetElement(unsigned int n, unsigned int m) const {
				arcs_assert(1<=n && n<=N && 1<=m && m<=M);	// サイズチェック
				return Data[n-1][m-1];
			}
			
			//! @brief すべての要素を指定した値で埋める関数
			//! @param[in] u 埋める値
			void FillAll(double u){
				for(unsigned int i=0;i<N;i++){
					for(unsigned int j=0;j<M;j++) Data[i][j] = u;
				}
			}
			
			//! @brief すべての要素を指定したゼロで埋める関数
			void FillAllZero(void){
				FillAll(0);
			}

			//! @brief 行列の幅(列数)を返す関数
			//! @return 行列の幅
			unsigned int GetWidthLength(void) const {
				return N;
			}

			//! @brief 行列の高さ(行数)を返す関数
			//! @return 行列の幅
			unsigned int GetHeightLength(void) const {
				return M;
			}
			
			//! @brief MatrixクラスがSIMD命令が有効に設定されているかを返す関数
			//! @return SIMD有効 = true, SIMD無効 = false
			bool isEnabledSIMD(void){
				return false;
			}
			
			//! @brief 行列のサイズの表示 (この関数はマクロを介して呼ばれることを想定している)
			//! @param[in] u		表示する行列
			//! @param[in] varname	変数名
			friend void PrintMatSize_Macro(const Matrix& u, const std::string& varname){
				printf("%s = [ %u x %u ]\n",varname.c_str(),u.N,u.M);
			}
			
			//! @brief 行列の要素を表示 (この関数はマクロを介して呼ばれることを想定している)
			//! @param[in] u		表示する行列
			//! @param[in] format	表示形式 (%1.3e とか %5.3f とか printfと同じ)
			//! @param[in] varname	変数名
			friend void PrintMatrix_Macro(const Matrix& u, const std::string& format, const std::string& varname){
				printf("%s = \n", varname.c_str());
				for(unsigned int j=0;j<u.M;j++){
					printf("[ ");
					for(unsigned int i=0;i<u.N;i++){
						printf(format.c_str(), u.Data[i][j]);
						printf(" ");
					}
					printf("]\n");
				}
				printf("\n");
			}
			
			//! @brief m行n列の零行列を返す関数
			//! @return 零行列
			static constexpr Matrix zeros(void){
				Matrix ret;
				return ret;
			}
			
			//! @brief m行n列の要素がすべて1の行列を返す関数
			//! @return 1行列
			static constexpr Matrix ones(void){
				Matrix ret;
				ret.FillAll(1);
				return ret;
			}
			
			//! @brief n行n列の単位行列を返す関数
			//! @return 単位行列
			static constexpr Matrix ident(void){
				static_assert(N == M, "Matrix Size Error");	// 行列のサイズチェック
				Matrix ret;
				for(unsigned int i=0;i<N;i++) ret.SetElem(i+1,i+1,1);	// 対角成分を1で埋める
				return ret;
			}
			
			//! @brief n行n列の単位行列を返す関数 (identのエイリアス)
			//! @return 単位行列
			static constexpr Matrix eye(void){
				return ident();
			}
			
			//! @brief 転置行列を返す関数
			//! @param[in]	U	入力行列
			//! @return	結果
			friend Matrix<MM,NN> tp(const Matrix& U){
				Matrix<U.M,U.N> y;
				for(unsigned int j=0;j<U.M;j++){
					for(unsigned int i=0;i<U.N;i++) y.Data[j][i] = U.Data[i][j];
				}
				return y;
			}
			
			//! @brief 行列のトレースを返す関数
			//! @param[in]	U	入力行列
			//! @return	結果
			friend double tr(const Matrix& U){
				static_assert(U.N == U.M, "Matrix Size Error");	// 正方行列であることのチェック
				double y = 0;
				for(unsigned int i=0;i<U.N;i++) y += (double)U.Data[i][i];
				return y;
			}
			
			//! @brief 行列の対角要素の総積を返す関数
			//! @param[in]	U	入力行列
			//! @return	結果
			friend double prod(const Matrix& U){
				static_assert(U.N == U.M, "Matrix Size Error");	// 正方行列であることのチェック
				double y = 1;
				for(unsigned int i=0;i<U.N;i++) y *= U.Data[i][i];
				return y;
			}
			
			//! @brief 行列の対角要素を返す関数
			//! @param[in]	U	入力行列
			//! @return	結果
			friend Matrix<NN,1> diag(const Matrix& U){
				static_assert(U.N == U.M, "Matrix Size Error");	// 正方行列であることのチェック
				Matrix<U.N,1> y;
				for(unsigned int i=0;i<U.N;i++) y.Data[i][0] = U.Data[i][i];
				return y;
			}
			
			//! @brief 行列要素の絶対値を返す関数
			//! @param[in]	U	入力行列
			//! @return	結果
			friend Matrix absm(const Matrix& U){
				Matrix Y;
				for(unsigned int i=0;i<U.N;i++){
					for(unsigned int j=0;j<U.N;j++) Y.Data[i][j] = fabs(U.Data[i][j]);
				}
				return Y;
			}
			
			//! @brief 行方向(横方向)へ加算して列ベクトルを出力する関数
			//! @param[in]	U	入力行列
			//! @return	結果
			friend Matrix<1,MM> sumrow(const Matrix& U){
				Matrix<1,U.M> y;
				for(unsigned int i=0;i<U.N;i++){
					for(unsigned int j=0;j<U.M;j++) y.Data[0][j] += U.Data[i][j];
				}
				return y;
			}
			
			//! @brief 列方向(縦方向)へ加算して行ベクトルを出力する関数
			//! @param[in]	U	入力行列
			//! @return	結果
			friend Matrix<NN,1> sumcolumn(const Matrix& U){
				Matrix<U.N,1> y;
				for(unsigned int i=0;i<U.N;i++){
					for(unsigned int j=0;j<U.M;j++) y.Data[i][0] += U.Data[i][j];
				}
				return y;
			}
			
			//! @brief ベクトル要素の最大値を返す関数
			//! @param[in]	u	入力ベクトル
			//! @return	結果
			friend double max(const Matrix& u){
				static_assert( (u.N == 1)||(u.M == 1), "Input is NOT vector." );	// ベクトルかチェック
				unsigned int k = 0;
				double y = 0;
				if(u.M == 1){
					// 行ベクトルの場合
					for(unsigned int i=1;i<u.N;i++)if(u.Data[k][0] < u.Data[i][0]) k = i;
					y = u.Data[k][0];
				}
				if(u.N == 1){
					// 列ベクトルの場合
					for(unsigned int i=1;i<u.M;i++)if(u.Data[0][k] < u.Data[0][i]) k = i;
					y = u.Data[0][k];
				}
				return y;
			}
			
			//! @brief ベクトル要素の絶対値の最大値を返す関数
			//! @param[in]	u	入力ベクトル
			//! @return	結果
			friend double absmax(const Matrix& u){
				static_assert( (u.N == 1)||(u.M == 1), "Input is NOT vector." );	// ベクトルかチェック
				unsigned int k = 0;
				double y = 0;
				if(u.M == 1){
					// 行ベクトルの場合
					for(unsigned int i=1;i<u.N;i++)if(fabs(u.Data[k][0]) < fabs(u.Data[i][0])) k = i;
					y = u.Data[k][0];
				}
				if(u.N == 1){
					// 列ベクトルの場合
					for(unsigned int i=1;i<u.M;i++)if(fabs(u.Data[0][k]) < fabs(u.Data[0][i])) k = i;
					y = u.Data[0][k];
				}
				return y;
			}
			
			//! @brief ベクトル要素の最大値の要素番号を返す関数
			//! @param[in]	u	入力ベクトル
			//! @return	結果
			friend unsigned int maxidx(const Matrix& u){
				static_assert( (u.N == 1)||(u.M == 1), "Input is NOT vector." );	// ベクトルかチェック
				unsigned int k = 0;
				if(u.M == 1){
					// 行ベクトルの場合
					for(unsigned int i=1;i<u.N;i++)if(u.Data[k][0] < u.Data[i][0]) k = i;
				}
				if(u.N == 1){
					// 列ベクトルの場合
					for(unsigned int i=1;i<u.M;i++)if(u.Data[0][k] < u.Data[0][i]) k = i;
				}
				return k + 1;
			}
			
			//! @brief ベクトル要素の絶対値の最大値の要素番号を返す関数
			//! @param[in]	u	入力ベクトル
			//! @return	結果
			friend unsigned int absmaxidx(const Matrix& u){
				static_assert( (u.N == 1)||(u.M == 1), "Input is NOT vector." );	// ベクトルかチェック
				unsigned int k = 0;
				if(u.M == 1){
					// 行ベクトルの場合
					for(unsigned int i=1;i<u.N;i++)if(fabs(u.Data[k][0]) < fabs(u.Data[i][0])) k = i;
				}
				if(u.N == 1){
					// 列ベクトルの場合
					for(unsigned int i=1;i<u.M;i++)if(fabs(u.Data[0][k]) < fabs(u.Data[0][i])) k = i;
				}
				return k + 1;
			}
			
			//! @brief 行列の無限大ノルムを返す関数
			//! @param[in]	U	入力行列
			//! @return	結果
			friend double infnorm(const Matrix& U){
				return max(sumcolumn(absm(tp(U))));
			}
			
			//! @brief 指定した行から横ベクトルとして抽出する関数
			//! @param[in]	U	入力行列
			//! @param[in]	m	抽出したい行
			//! @return	結果
			friend Matrix<NN,1> getrow(const Matrix& U, unsigned int m){
				arcs_assert(0 < m && m <= U.M);	// 指定した行が行列の高さ以内かチェック
				Matrix<U.N,1> y;
				for(unsigned int i=0;i<U.N;i++) y.Data[i][0] = U.Data[i][m-1];
				return y;
			}
			
			//! @brief 指定した行を横ベクトルで上書きする関数
			//! @param[in,out]	U	入出力行列
			//! @param[in]	v	横ベクトル
			//! @param[in]	m	上書きしたい行
			friend void setrow(Matrix& U, const Matrix<NN,1>& v, unsigned int m){
				static_assert(U.N == v.N, "Matrix Size Error");	// 行列の幅と横ベクトルの長さが等しいかチェック
				static_assert(v.M == 1, "Input is NOT vector.");// 行ベクトルかチェック
				arcs_assert(0 < m && m <= U.M);	// 指定した行が行列の高さ以内かチェック
				for(unsigned int i=0;i<U.N;i++) U.Data[i][m-1] = v.Data[i][0];
			}
			
			//! @brief 指定した行を横ベクトル(std::array)で上書きする関数
			//! @param[in,out]	U	入出力行列
			//! @param[in]	v	横ベクトル
			//! @param[in]	m	上書きしたい行
			friend void setrow(Matrix& U, const std::array<double, NN>& v, unsigned int m){
				static_assert(U.N == NN, "Matrix Size Error");	// 行列の幅と横ベクトルの長さが等しいかチェック
				arcs_assert(0 < m && m <= U.M);	// 指定した行が行列の高さ以内かチェック
				for(unsigned int i=0;i<U.N;i++) U.Data[i][m-1] = v.at(i);
			}
			
			//! @brief 指定した行と行を入れ替える関数
			//! @param[in,out]	U	入出力行列
			//! @param[in]	m1	指定行1
			//! @param[in]	m2	指定行2
			friend void swaprow(Matrix& U, unsigned int m1, unsigned int m2){
				arcs_assert(m1 <= U.M);	// 行が高さ以内かチェック
				arcs_assert(m2 <= U.M);	// 行が高さ以内かチェック
				Matrix<U.N,1> p, q;	// バッファ用ベクトル
				p = getrow(U,m1);	// m1行目を抽出
				q = getrow(U,m2);	// m2行目を抽出
				setrow(U, p, m2);	// m2行目に書き込み
				setrow(U, q, m1);	// m1行目に書き込み
			}
			
			//! @brief m行目のn1列目からm2列目を数値aで埋める関数  (n1 <= n2 であること)
			//! @param[in,out]	U	入出力行列
			//! @param[in]	a	埋める値
			//! @param[in]	m	指定行
			//! @param[in]	n1	開始列
			//! @param[in]	n2	終了列
			friend void fillrow(Matrix& U, double a, unsigned int m, unsigned int n1, unsigned int n2){
				arcs_assert(0 < m && m <= U.M);		// 列が幅以内かチェック
				arcs_assert(0 < n1 && n1 <= U.N);	// 列が高さ以内かチェック
				arcs_assert(0 < n2 && n2 <= U.N);	// 列が高さ以内かチェック
				arcs_assert(n1 <= n2);				// 開始列と終了列が入れ替わらないかチェック
				for(unsigned int i=n1;i<=n2;i++) U.Data[i-1][m-1] = a;
			}
			
			//! @brief 指定した列から縦ベクトルとして抽出する関数
			//! @param[in]	U	入力行列
			//! @param[in]	n	指定列
			//! @return	結果
			friend Matrix<1,MM> getcolumn(const Matrix& U, unsigned int n){
				arcs_assert(n <= U.N);	// 指定した列が行列の幅以内かチェック
				Matrix<1,U.M> y;
				for(unsigned int i=0;i<U.M;i++) y.Data[0][i] = U.Data[n-1][i];
				return y;
			}
			
			//! @brief 指定した列を列ベクトルで上書きする関数
			//! @param[in,out]	U	入出力行列
			//! @param[in]	v	縦ベクトル
			//! @param[in]	n	上書きしたい列
			friend void setcolumn(Matrix& U, const Matrix<1,MM>& v, unsigned int n){
				static_assert(U.M == v.M, "Matrix Size Error");		// 行列の高さ列ベクトルの長さが等しいかチェック
				static_assert(v.N == 1, "Input is NOT vector.");	// 列ベクトルかチェック
				arcs_assert(0 < n && n <= U.N);	// 指定した列が行列の幅以内かチェック
				for(unsigned int i=0;i<U.M;i++) U.Data[n-1][i] = v.Data[0][i];
			}
			
			//! @brief 指定した列を列ベクトル(std::array)で上書きする関数
			//! @param[in,out]	U	入出力行列
			//! @param[in]	v	縦ベクトル
			//! @param[in]	n	上書きしたい列
			friend void setcolumn(Matrix& U, const std::array<double,MM>& v, unsigned int n){
				static_assert(U.M == MM, "Matrix Size Error");		// 行列の高さ列ベクトルの長さが等しいかチェック
				arcs_assert(0 < n && n <= U.N);	// 指定した列が行列の幅以内かチェック
				for(unsigned int i=0;i<U.M;i++) U.Data[n-1][i] = v.at(i);
			}
			
			//! @brief 指定した列と列を入れ替える関数
			//! @param[in,out]	U	入出力行列
			//! @param[in]	n1	指定列1
			//! @param[in]	n2	指定列2
			friend void swapcolumn(Matrix& U, unsigned int n1, unsigned int n2){
				arcs_assert(n1 <= U.N);	// 列が幅以内かチェック
				arcs_assert(n2 <= U.N);	// 列が幅以内かチェック
				Matrix<1,U.M> p, q;	// バッファ用ベクトル
				p = getcolumn(U,n1);	// n1列目を抽出
				q = getcolumn(U,n2);	// n2列目を抽出
				setcolumn(U,p,n2);		// n2列目に書き込み
				setcolumn(U,q,n1);		// n1列目に書き込み
			}
			
			//! @brief n列目のm1行目からm2行目を数値aで埋める関数  (m1 <= m2 であること)
			//! @param[in,out]	U	入出力行列
			//! @param[in]	a	埋める値
			//! @param[in]	n	指定列
			//! @param[in]	m1	開始行
			//! @param[in]	m2	終了行
			friend void fillcolumn(Matrix& U, double a, unsigned int n, unsigned int m1, unsigned int m2){
				arcs_assert(n <= U.N);	// 列が幅以内かチェック
				arcs_assert(m1 <= U.M);	// 行が高さ以内かチェック
				arcs_assert(m2 <= U.M);	// 行が高さ以内かチェック
				arcs_assert(m1 <= m2);	// 開始行と終了行が入れ替わらないかチェック
				for(unsigned int i=m1;i<=m2;i++) U.Data[n-1][i-1] = a;
			}
			
			//! @brief 並び替え記憶列ベクトルvの行番号に従って，入力行列Uの行を並び替える関数
			//! @param[in]	U	入力行列
			//! @param[in]	v	並び替え記憶列ベクトル
			//! @return	結果
			friend Matrix orderrow(const Matrix& U, const Matrix<1,MM>& v){
				static_assert(U.M == v.M, "Matrix and Vector Size Error");	// 行列の高さと列ベクトルの長さが同じかチェック
				Matrix Y;
				for(unsigned int i=0;i<U.M;i++){
					setrow(Y,  getrow(U, (unsigned int)v.Data[0][i]), i+1);
				}
				return Y;
			}
			
			//! @brief 並び替え記憶列ベクトルvが昇順になるように，入力行列Uの行を並び替えて元に戻す関数
			//! @param[in]	U	入力行列
			//! @param[in]	v	並び替え記憶列ベクトル
			//! @return	結果
			friend Matrix reorderrow(const Matrix& U, const Matrix<1,MM>& v){
				static_assert(U.M == v.M, "Matrix and Vector Size Error");	// 行列の高さと列ベクトルの長さが同じかチェック
				Matrix Y = U;
				Matrix<1,MM> p = v;
				for(unsigned int i=1;i<=U.M;i++){
					swaprow(Y,i,p.Data[0][i-1]);
					swaprow(p,i,p.Data[0][i-1]);
				}
				return Y;
			}
			
			//! @brief LU分解
			//! @param[in]	A	入力行列
			//! @param[out]	L	下三角行列
			//! @param[out]	U	上三角行列
			//! @param[out]	v	並べ替え記憶列ベクトル
			friend enum LUperm LU(const Matrix& A, Matrix& L, Matrix& U, Matrix<1,MM>& v){
				static_assert(A.N == A.M, "Matrix Size Error");	// 正方行列かチェック
				Matrix X = A;
				unsigned int i,j,k,l,perm_count = 0;
				double max_buff;
				Matrix::LUperm ret;
				
				// 並べ替え記憶列ベクトルの準備 (本当はvベクトルは整数型にすべき。テンプレート化のときに修正する予定)
				for(i=0;i<X.N;i++)v.Data[0][i] = i + 1;	// 並べ替え記憶列ベクトルに1から昇順の番号を書き込む
				
				for(j=0;j<X.N-1;j++){
					// j列目の中での行要素の最大値を探す
					k = j;	// 対角要素番号で初期化
					max_buff = fabs(X.Data[j][j]);	// 対角要素で初期化
					for(i=j+1;i<X.M;i++){
						if(max_buff < fabs(X.Data[j][i])){
							k = i;
							max_buff = fabs(X.Data[j][i]);
						}
					}
					// 行入れ替えが必要なときは，
					if(k != j){
						swaprow(v,j+1,k+1);				// 並べ替え記憶列ベクトルのj行目とk行目を並べ替え
						swaprow(X,j+1,k+1);				// j行目とk行目を並べ替え
						perm_count++;
					}
					if( fabs(X.Data[j][j]) < X.epsilon )continue;
					{
						// 対角要素が零なら，j列目においてはLU分解が終わっているので，以下の処理はスキップ
						// ここからLU分解
						for(i=j+1;i<X.M;i++){
							X.Data[j][i] /= X.Data[j][j];	// 対角要素で除算
							for(l=j+1;l<X.N;l++){
								X.Data[l][i] -= X.Data[j][i]*X.Data[l][j];
							}
						}
					}
				}
				// 下三角行列と上三角行列に分離する
				for(j=0;j<X.N;j++){
					for(i=j;i<X.M;i++){
						if(i==j){
							L.Data[j][i] = 1;				// 下三角行列の対角要素はすべて1
						}else{
							L.Data[j][i] = X.Data[j][i];	// 下三角のみコピー
						}
					}
					for(i=0;i<=j;i++) U.Data[j][i] = X.Data[j][i];		// 上三角のみコピー
				}
				// 並べ替え回数の判定
				if(perm_count % 2 == 0){
					ret = Matrix::EVEN;	// 奇数のとき
				}else{
					ret = Matrix::ODD;	// 偶数のとき
				}
				return ret;
			}
			
			//! @brief Ax = bの形の線形連立1次方程式をxについて解く関数
			//! @param[in]	A	係数行列
			//! @param[in]	b	係数ベクトル
			//! @return	解ベクトル
			friend Matrix<1,MM> solve(const Matrix& A, const Matrix<1,MM>& b){
				static_assert(A.N == A.M, "Matrix Size Error");	// Aが正方行列かチェック
				static_assert(b.M == A.M, "Matrix and Vector Size Error");	// Aの高さとbの高さが同じかチェック
				static_assert(b.N == 1, "Input is NOT vector.");	// bは縦ベクトルかチェック
				Matrix<A.N,A.N> L, U;
				Matrix<1,A.N> d, x, v, bb;
				unsigned int i,j;
				double buff=0;
				// Ax = b において A をLU分解すると，(LU)x = b になって L(Ux) = b と表現できることを利用する。
				LU(A,L,U,v);		// まず，LU分解(並べ替え有り)してから，Ux = d と勝手に置き換えて考える。
				bb = orderrow(b,v);	// bベクトルも並べ替える
				// その次に Ld = b の方を d について解いて，
				// (下記では，Lの対角要素が1となるLU分解がなされているものとして計算する)
				d.Data[0][0] = bb.Data[0][0];
				for(i=1;i<A.N;i++){
					for(j=0;j<=i-1;j++) buff += L.Data[j][i]*d.Data[0][j];
					d.Data[0][i] = (bb.Data[0][i] - buff);
					buff = 0;
				}
				// さらに Ux = d を x について解く。
				x.Data[0][A.N-1] = d.Data[0][A.N-1]/U.Data[A.N-1][A.N-1];
				for(int k=A.N-2;0<=k;k--){
					for(j=(unsigned int)k+1;j<A.N;j++){
						buff += U.Data[j][k]*x.Data[0][j];
					}
					x.Data[0][k] = (d.Data[0][k] - buff)/U.Data[k][k];
					buff = 0;
				}
				return x;	// 最終的な答えのxベクトルを返す
			}
			
			//! @brief 行列式の値を返す関数
			//! @param[in]	A	入力行列
			//! @return	結果
			friend double det(const Matrix& A){
				static_assert(A.N == A.M, "Matrix Size Error");	// Aが正方行列かチェック
				Matrix<A.N,A.N> L, U;
				Matrix<1,A.N> v;
				int sign;		// 符号
				if(LU(A,L,U,v) == Matrix::ODD){	// LU分解と符号判定
					sign = -1;	// 奇数のとき
				}else{
					sign =  1;	// 偶数のとき
				}
				// |A| = |L||U| でしかも |L|と|U|は対角要素の総積に等しく，さらにLの対角要素は1なので|L|は省略可。
				// 最後にLU分解のときの並べ替え回数によって符号反転をする。
				return (double)sign*prod(U);
			}
			
			//! @brief 逆行列を返す関数 (正則チェック無し)
			//! @param[in]	A	入力行列
			//! @return	結果
			friend Matrix inv(const Matrix& A){
				static_assert(A.N == A.M, "Matrix Size Error");	// Aが正方行列かチェック
				Matrix I = Matrix<A.N,A.N>::ident();	// 単位行列の生成
				Matrix<1,A.N> x, b;
				Matrix<A.N,A.N> Ainv;
				for(unsigned int n=1;n<=A.N;n++){
					b = getcolumn(I,n);		// 単位行列のn列目を切り出してbベクトルとする
					x = solve(A,b);			// Ax=bの連立1次方程式をxについて解く
					setcolumn(Ainv,x,n);	// xはAの逆行列のn列目となるので、Ainvのn列目にxを書き込む
				}
				return Ainv;	// 最終的に得られる逆行列を返す
			}
			
			//! @brief 逆行列を返す関数 (正則チェック有り)
			//! @param[in]	A	入力行列
			//! @return	結果
			friend Matrix inv_with_check(const Matrix& A){
				static_assert(A.N == A.M, "Matrix Size Error");	// Aが正方行列かチェック
				arcs_assert(A.epsilon < fabs(det(A)));			// 正則かチェック
				Matrix I = Matrix<A.N,A.N>::ident();	// 単位行列の生成
				Matrix<1,A.N> x, b;
				Matrix<A.N,A.N> Ainv;
				for(unsigned int n=1;n<=A.N;n++){
					b = getcolumn(I,n);		// 単位行列のn列目を切り出してbベクトルとする
					x = solve(A,b);			// Ax=bの連立1次方程式をxについて解く
					setcolumn(Ainv,x,n);	// xはAの逆行列のn列目となるので、Ainvのn列目にxを書き込む
				}
				return Ainv;	// 最終的に得られる逆行列を返す
			}
			
			//! @brief 左擬似逆行列を返す関数 (Aが縦長行列の場合)
			//! @param[in]	A	入力行列
			//! @return	結果
			friend Matrix<MM,NN> lpinv(const Matrix& A){
				static_assert(A.N < A.M, "Matrix Size Error");	// 縦長行列かチェック
				return inv(tp(A)*A)*tp(A);
			}

			//! @brief 右擬似逆行列を返す関数 (Aが横長行列の場合)
			//! @param[in]	A	入力行列
			//! @return	結果
			friend Matrix<MM,NN> rpinv(const Matrix& A){
				static_assert(A.M < A.N, "Matrix Size Error");	// 横長行列かチェック
				return tp(A)*inv(A*tp(A));
			}
			
			//! @brief 行列指数関数 e^(U)
			//! @param[in]	U	入力行列
			//! @param[in]	Order	パデ近似の次数
			//! @return	結果
			friend Matrix expm(const Matrix& U, unsigned int Order){
				static_assert(U.N == U.M, "Matrix Size Error");	// 正方行列かチェック
				unsigned int i;
				int e;
				double c = 1;
				bool flag = false;
				// ノルムでスケーリング
				frexp(infnorm(U),&e);
				Matrix<U.N,U.M> A;
				if(0 < e){
					A = pow(0.5,e+1)*U;
				}else{
					e = 0;
					A = 0.5*U;
				}
				// 行列のパデ近似の計算
				Matrix<A.N,A.N> I = ident();// 単位行列の生成
				Matrix<A.N,A.N> L = I, R = I, X = I, cX;
				for(i=1;i<=Order;i++){
					c = c*(double)(Order-i+1)/(double)(i*(2*Order - i+1));	// パデ近似係数の計算
					X = A*X;		// A^Mの計算
					cX = c*X;		// cM*A^Mの計算
					R += cX;		// R = I + c1*A + c2*A*A + c3*A*A*A + ... + cM*A^M
					if(flag == true){
						L += cX;	// L = I + c1*A + c2*A*A + c3*A*A*A + ... + cM*A^M の正の係数の場合
					}else{
						L -= cX;	// L = I + c1*A + c2*A*A + c3*A*A*A + ... + cM*A^M の負の係数の場合
					}
					flag = !flag;	// 正負係数の場合分け用フラグ
				}
				// スケールを元に戻す
				Matrix Y = inv(L)*R;
				for(i=0;i<(unsigned int)e+1;i++){
					Y = Y*Y;
				}
				return Y;	// 最終的に得られる行列指数を返す
			}
			
			//! @brief 指数行列の数値定積分[0,T]をする関数
			//! @param[in]	U	入力行列
			//! @param[in]	T	積分範囲の終わり
			//! @param[in]	N	分割数
			//! @param[in]	P	パデ近似の次数
			//! @return	結果
			friend Matrix integral_expm(const Matrix& U, const double T, const unsigned long DIV, const unsigned int P){
				static_assert(U.N == U.M, "Matrix Size Error");	// 正方行列かチェック
				const double h = T/((double)(2*DIV));	// 時間ステップ
				double t;								// 時刻
				// シンプソン法による定積分の実行
				Matrix<U.N,U.M> S1, S2;
				for(unsigned long i=1;i<=DIV;i++){
					t = h*(double)(2*i-1);
					S1 += expm(U*t,P);
				}
				for(unsigned long i=1;i<=DIV-1;i++){
					t = h*(double)(2*i);
					S2 += expm(U*t,P);
				}
				return h/3.0*( Matrix<U.N,U.N>::eye() + 4.0*S1 + 2.0*S2 + expm(U*T,P) );	// 最終的な定積分結果を返す
			}
			
			//! @brief 行列要素の指数関数を返す関数
			//! @param[in]	U	入力行列
			//! @return	結果
			friend Matrix expe(const Matrix& U){
				Matrix Y;
				for(unsigned int i=0;i<U.N;i++){
					for(unsigned int j=0;j<U.N;j++) Y.Data[i][j] = exp(U.Data[i][j]);
				}
				return Y;
			}
			
			//! @brief 行列要素の平方根を返す関数
			//! @param[in]	U	入力行列
			//! @return	結果
			friend Matrix sqrtm(const Matrix& U){
				Matrix Y;
				for(unsigned int i=0;i<U.N;i++){
					for(unsigned int j=0;j<U.N;j++) Y.Data[i][j] = sqrt(U.Data[i][j]);
				}
				return Y;
			}
			
			//! @brief 行列要素の平方根を参照で返す関数
			//! @param[in]	U	入力行列
			//! @param[out]	Y	結果
			friend void sqrtm(const Matrix& U, Matrix& Y){
				for(unsigned int i=0;i<U.N;i++){
					for(unsigned int j=0;j<U.N;j++) Y.Data[i][j] = sqrt(U.Data[i][j]);
				}
			}
			
		private:
			static constexpr double epsilon = 1e-12;	//!< 零とみなす閾値
			unsigned int Nindex;	//!< 横用カウンタ
			unsigned int Mindex;	//!< 縦用カウンタ
		public:
			static constexpr unsigned int N = NN;		//!< 行列の幅(列の数)
			static constexpr unsigned int M = MM;		//!< 行列の高さ(行の数)
			std::array<std::array<double, M>, N> Data;	//!< データ格納用変数 配列要素の順番は Data[N列][M行] なので注意
	};
	
}

#endif

