//! @file MotorSimulator.cc
//! @brief モータシミュレータ
//!
//! モータを模擬する
//!
//! @date 2020/03/16
//! @author Yokokura, Yuki
//
// Copyright (C) 2011-2030 Yokokura, Yuki
// This program is free software;
// you can redistribute it and/or modify it under the terms of the FreeBSD License.
// For details, see the License.txt file.

#include <cassert>
#include <tuple>
#include "MotorSimulator.hh"
#include "Discret.hh"

// ARCS組込み用マクロ
#ifdef ARCS_IN
	// ARCSに組み込まれる場合
	#include "ARCSassert.hh"
	#include "ARCSeventlog.hh"
#else
	// ARCSに組み込まれない場合
	#define arcs_assert(a) (assert(a))
	#define PassedLog()
	#define EventLog(a)
	#define EventLogVar(a)
#endif

using namespace ARCS;

//! @brief コンストラクタ
//! @param[in]	TrqConst	[Nm/A] トルク定数
//! @param[in]	MotorInert	[kgm^2]モータ慣性
//! @param[in]	MotorFric	[Nm/(rad/s)]モータ粘性
//! @param[in]	SmplTime	[s] サンプリング時間
MotorSimulator::MotorSimulator(const double TrqConst, const double MotorInert, const double MotorFric, const double SmplTime)
	: Kt(TrqConst), Jm(MotorInert), Dm(MotorFric), Ts(SmplTime), iq(0), taul(0),
	  Ac(), Bc(), Ad(), Bd(), u(), x(), x_next()
{
	// 連続系A行列の設定
	Ac.Set(
		-Dm/Jm,  0,
		     1,  0
	);
	
	// 連続系B行列の設定
	Bc.Set(
		Kt/Jm, -1.0/Jm,
		    0,       0
	);
	
	// 離散化
	std::tie(Ad,Bd) = Discret::GetDiscSystem(Ac,Bc,Ts);
	
	PassedLog();
}

//! @brief ムーブコンストラクタ
MotorSimulator::MotorSimulator(MotorSimulator&& r)
	: Kt(r.Kt), Jm(r.Jm), Dm(r.Dm), Ts(r.Ts), iq(r.iq), taul(r.taul),
	  Ac(r.Ac), Bc(r.Bc), Ad(r.Ad), Bd(r.Bd), u(r.u), x(r.x), x_next(r.x_next)
{
	
}

//! @brief デストラクタ
MotorSimulator::~MotorSimulator(){
	PassedLog();
}

//! @brief モータ速度と位置を返す関数
//! @param[in]	current	[A] 電流
//! @param[in]	loadtorque	[Nm] 負荷トルク
//! @return	モータ速度 [rad/s], 位置 [rad]
std::tuple<double,double> MotorSimulator::GetResponses(const double current, const double loadtorque){
	double velocity, position;
	
	// 入力ベクトルの設定
	u.Set(
		current,
		loadtorque
	);
	
	// 離散系状態方程式の計算
	x_next = Ad*x + Bd*u;
	
	// 状態ベクトルの更新
	x = x_next;
	
	// 状態ベクトルから抽出
	velocity = x.GetElement(1,1);
	position = x.GetElement(1,2);
	
	// 速度と位置を返す
	return std::forward_as_tuple(velocity, position);
}

//! @brief モータ電流と負荷トルクを設定する関数
//! @param[in]	current	[A] 電流
//! @param[in]	loadtorque	[Nm] 負荷トルク
void MotorSimulator::SetCurrentAndLoadTorque(const double current, const double loadtorque){
	iq = current;
	taul = loadtorque;
}

//! @brief 予めセットされたモータ電流と負荷トルクから速度と位置を返す関数
//! @return モータ速度 [rad/s]，位置 [rad]
std::tuple<double, double> MotorSimulator::GetResponses(void){
	return GetResponses(iq, taul);
}

//! @brief モータ慣性を設定する関数
void MotorSimulator::SetMotorInertia(const double inertia){
	Jm = inertia;
}

//! @brief 状態ベクトルをリセットする関数
void MotorSimulator::ClearStateVector(void){
	x = Matrix<1,2>::zeros();
	x_next = Matrix<1,2>::zeros();
}

