//! @file MultiLayerPerceptron3.hh
//! @brief 多層パーセプトロンクラス3層版
//!
//! 3層ニューラルネットワーク
//!
//! @date 2020/03/05
//! @author Yuki YOKOKURA
//
// Copyright (C) 2011-2020 Yuki YOKOKURA
// This program is free software;
// you can redistribute it and/or modify it under the terms of the FreeBSD License.
// For details, see the License.txt file.

#ifndef MULTILAYERPERCEPTRON3
#define MULTILAYERPERCEPTRON3

#include <cassert>
#include <array>
#include <string>
#include "Matrix.hh"
#include "SingleLayerPerceptron.hh"
#include "FrameGraphics.hh"
#include "CuiPlot.hh"

// ARCS組込み用マクロ
#ifdef ARCS_IN
	// ARCSに組み込まれる場合
	#include "ARCSassert.hh"
	#include "ARCSeventlog.hh"
#else
	// ARCSに組み込まれない場合
	#define arcs_assert(a) (assert(a))
	#define PassedLog()
	#define EventLog(a)
	#define EventLogVar(a)
#endif

namespace ARCS {	// ARCS名前空間
	//! @brief エポック毎にデータセットをシャッフルするかどうか
	enum class MlpShuffle {
		ENABLE,	//!< エポック毎にシャッフルする
		DISABLE	//!< エポック毎にシャッフルしない
	};
	
	//! @brief 多層パーセプトロンクラス3層版
	// 下記のテンプレートの実装はとてもアホっぽいが，
	// C++17ではテンプレートパラメータにローカルなstructも配列も渡せないし，
	// パラメータパックは展開めんどくせーし，という言い訳をここに書いておく
	template <
		size_t InSize,		// 入力層のサイズ
		size_t HidSize,		// 隠れ層のサイズ
		size_t OutSize,		// 出力層のサイズ 
		size_t MinbatSize,	// ミニバッチサイズ
		size_t MinbatNum,	// ミニバッチ数
		size_t Epoch,		// エポック数
		ActvFunc InFunc,			// 入力層の活性化関数
		ActvFunc HidFunc,			// 隠れ層の活性化関数
		ActvFunc OutFunc,			// 出力層の活性化関数
		SlpDescentTypes GradDesType	// 勾配降下法のタイプ
	>
	class MultiLayerPerceptron3 {
		public:
			//! @brief コンストラクタ
			MultiLayerPerceptron3()
				: InputLayer(), HiddenLayer(), OutputLayer(),
				  X(), ZIn(), ZHid(), Y(), D(), WDeltaOut(), WDeltaHid(), WDeltaIn(),
				  EpochNumbers(), TrainingError(), TestError(),
				  XTest(), ZInTest(), ZHidTest(), YTest(), DTest(),
				  zin(), zhid()
			{
				
			}
			
			//! @brief ムーブコンストラクタ
			//! @param[in]	r	右辺値
			MultiLayerPerceptron3(MultiLayerPerceptron3&& r)
				: InputLayer(), HiddenLayer(), OutputLayer(),
				  X(), ZIn(), ZHid(), Y(), D(), WDeltaOut(), WDeltaHid(), WDeltaIn(),
				  EpochNumbers(), TrainingError(), TestError(),
				  XTest(), ZInTest(), ZHidTest(), YTest(), DTest(),
				  zin(), zhid()
			{
				
			}
			
			//! @brief デストラクタ
			~MultiLayerPerceptron3(){
				
			}
			
			//! @brief 重み行列の初期化
			void InitWeight(void){
				InputLayer.template  InitWeight<SlpInitTypes::HE>(InSize);	// 入力層の重み行列の乱数による初期化
				HiddenLayer.template InitWeight<SlpInitTypes::HE>(InSize);	// 内部層の重み行列の乱数による初期化
				OutputLayer.template InitWeight<SlpInitTypes::HE>(HidSize);	// 出力層の重み行列の乱数による初期化
			}
			
			//! @brief モーメンタム確率的勾配降下法の更新ゲイン(学習率)の設定
			//! @param[in]	epsilon	更新ゲイン(学習率)
			//! @param[in]	alpha	更新ゲイン(運動量項の学習率)
			void SetGainOfMomentumSGD(double epsilon, double alpha){
				InputLayer.SetGainOfMomentumSGD(epsilon, alpha);
				HiddenLayer.SetGainOfMomentumSGD(epsilon, alpha);
				OutputLayer.SetGainOfMomentumSGD(epsilon, alpha);
			}
			
			//! @brief ドロップアウト率の設定
			//! @param[in]	DropoutRate	ドロップアウト率(1のときドロップアウトしない，0のとき全部ドロップアウト)
			void SetDropoutRate(double DropoutRate){
				InputLayer.SetDropoutRate(DropoutRate);
				HiddenLayer.SetDropoutRate(DropoutRate);
				OutputLayer.SetDropoutRate(DropoutRate);
			}
			
			//! @brief 誤差逆伝播法による訓練をする関数
			//! @tparam	ED	エポックループ表示数
			//! @tparam	DS	データセットクラスの名前
			//! @tparam	MS	エポック毎にシャッフルするかどうかの設定
			template<size_t ED, typename DS, MlpShuffle MS>
			void Train(DS& DataSets){
				InitWeight();				// 重み行列の初期化
				InputLayer.NomalizeDataset(DataSets.IrisMeas);	// データセットの正規化(標準化)
				DataSets.template GetMeasurementData<MinbatSize, InSize>(XTest, 14);	// ミニバッチ分のテスト入力データを取得
				DataSets.template GetClassificationData<MinbatSize, OutSize>(DTest, 14);// ミニバッチ分のテスト目標データを取得
				
				// 誤差逆伝播法のエポック数分のループ
				printf("Epoch : Train Error   , Test Error\n");
				for(size_t i = 0; i < Epoch; ++i){
					// ミニバッチ数分のループ
					for(size_t j = 0; j < MinbatNum; ++j){
						// 訓練データの準備
						DataSets.template GetMeasurementData<MinbatSize, InSize>(X, j + 1);		// ミニバッチ分の入力データを取得
						DataSets.template GetClassificationData<MinbatSize, OutSize>(D, j + 1);	// ミニバッチ分の目標データを取得
						
						// ドロップアウトの準備
						InputLayer.CalcDropout();
						HiddenLayer.CalcDropout();
						
						// 順伝播計算
						InputLayer.template CalcForward<InDrop>(X, ZIn);		// 入力層の順伝播計算
						HiddenLayer.template CalcForward<HidDrop>(ZIn, ZHid);	// 内部層の順伝播計算
						OutputLayer.CalcForward(ZHid, Y);						// 出力層の順伝播計算
						
						// 逆伝播による誤差行列計算
						OutputLayer.CalcDeltaForOutputLayer(Y, D, WDeltaOut);			// 出力層の逆伝播計算
						HiddenLayer.template CalcDelta<HidDrop>(WDeltaOut, WDeltaHid);	// 内部層の逆伝播計算
						InputLayer.template CalcDelta<InDrop>(WDeltaHid, WDeltaIn);		// 入力層の逆伝播計算
						
						// 重み行列の更新
						InputLayer.template  UpdateWeight<GradDesType>(X);		// 入力層の重み更新
						HiddenLayer.template UpdateWeight<GradDesType>(ZIn);	// 内部層の重み更新
						OutputLayer.template UpdateWeight<GradDesType>(ZHid);	// 出力層の重み更新
					}
					if(MS == MlpShuffle::ENABLE) DataSets.ShuffleDataSets();		// データセットのシャッフル
					
					// テストデータの順伝播計算
					InputLayer.template CalcForward<InDropEst>(XTest, ZInTest);		// 入力層の順伝播計算
					HiddenLayer.template CalcForward<HidDropEst>(ZInTest, ZHidTest);// 内部層の順伝播計算
					OutputLayer.CalcForward(ZHidTest, YTest);						// 出力層の順伝播計算
					
					// 学習曲線評価用のデータ保持
					EpochNumbers[i] = i;
					TrainingError[i] = 10.0*log10(OutputLayer.GetTrainingError());			// 訓練誤差
					TestError[i] = 10.0*log10(OutputLayer.GetTestError(YTest, DTest));		// テスト誤差
					
					// エポック毎の訓練誤差とテスト誤差の表示
					if(i % (Epoch/ED) == 0)
						printf("%5lu : % 10.12f, % 10.12f\n", i, TrainingError[i], TestError[i]);
				}
			}
			
			//! @brief 重み行列の表示
			void DispWeight(void){
				InputLayer.DispWeight();	// 入力層の重みの表示
				HiddenLayer.DispWeight();	// 内部層の重みの表示
				OutputLayer.DispWeight();	// 出力層の重みの表示
			}
			
			//! @brief バイアスベクトルの表示
			void DispBias(void){
				InputLayer.DispBias();
				HiddenLayer.DispBias();
				OutputLayer.DispBias();
			}
			
			//! @brief パーセプトロン設定値の表示
			void DispSettings(void){
				InputLayer.DispSettings();
				HiddenLayer.DispSettings();
				OutputLayer.DispSettings();
			}
			
			//! @brief 最後の学習結果を使った順伝播出力値と正解値の確認
			void DispFinalTestData(void){
				PrintMatrix(YTest, "%1.0f");	// 順伝播出力値
				PrintMatrix(DTest, "%1.0f");	// 正解値
			}
			
			//! @brief 学習曲線グラフのPNG画像ファイルを書き出す関数
			//! @tparam	YMIN	グラフの縦軸最小値 [dB]
			//! @tparam YMAX	グラフの縦軸最大値 [dB]
			//! @param[in]	FileName	ファイル名
			template<int YMIN, int YMAX>
			void WriteLeaningCurvePNG(std::string FileName){
				// 学習曲線のグラフの表示
				FrameGraphics FG(1000, 500);
				CuiPlot Plot(FG, 0,0,1000,500);
				Plot.SetAxisLabels("Epoch Number", "Errors 10log10 [dB]");
				Plot.SetRanges(0, Epoch, (double)YMIN, (double)YMAX);
				Plot.SetGridLabelFormat("%5.0f", "%2.1f");
				Plot.DrawAxis();
				Plot.DrawLegend(1, "Training", FGcolors::CYAN);
				Plot.DrawLegend(2, "Test", FGcolors::MAGENTA);
				Plot.Plot(EpochNumbers, TrainingError, CuiPlotTypes::PLOT_BOLDSTAIRS, FGcolors::CYAN);
				Plot.Plot(EpochNumbers, TestError, CuiPlotTypes::PLOT_BOLDSTAIRS, FGcolors::MAGENTA);
				FG.SavePngImageFile(FileName);
			}
			
			//! @brief 訓練済みニューラルネットワークを使った推定計算
			//! @param[in]	x	ニューラルネットワークの入力ベクトル
			//! @param[out]	y	ニューラルネットワークの出力ベクトル
			void Estimate(const Matrix<1,InSize>& x, Matrix<1,OutSize>& y){
				// 順伝播計算
				Matrix<1,InSize> xn = x;			// 正規化用
				InputLayer.NormalizeInput(xn);		// 入力データの正規化(標準化)
				InputLayer.template CalcForward<InDropEst>(xn, zin);	// 入力層の順伝播計算
				HiddenLayer.template CalcForward<HidDropEst>(zin, zhid);// 内部層の順伝播計算
				OutputLayer.CalcForward(zhid, y);						// 出力層の順伝播計算
			}
			
			//! @brief 各レイヤの重み行列とバイアスベクトルをCSVファイルとして保存する関数
			//! @param[in]	FileName	CSVファイル名(拡張子なし)
			void SaveWeightAndBias(const std::string& FileName){
				InputLayer.SaveWeightAndBias(FileName + "-InputLayerW.csv", FileName + "-InputLayerb.csv");
				HiddenLayer.SaveWeightAndBias(FileName + "-HiddenLayerW.csv", FileName + "-HiddenLayerb.csv");
				OutputLayer.SaveWeightAndBias(FileName + "-OutputLayerW.csv", FileName + "-OutputLayerb.csv");
			}
			
			//! @brief CSVファイルから各レイヤの重み行列とバイアスベクトルに読み込む関数
			//! @param[in]	FileName	CSVファイル名(拡張子なし)
			void LoadWeightAndBias(const std::string& FileName){
				InputLayer.LoadWeightAndBias(FileName + "-InputLayerW.csv", FileName + "-InputLayerb.csv");
				HiddenLayer.LoadWeightAndBias(FileName + "-HiddenLayerW.csv", FileName + "-HiddenLayerb.csv");
				OutputLayer.LoadWeightAndBias(FileName + "-OutputLayerW.csv", FileName + "-OutputLayerb.csv");
			}
			
			//! @brief 各レイヤのパーセプトロンの設定値をCSVファイルとして保存する関数
			//! @param[in]	FileName	CSVファイル名(拡張子なし)
			void SaveSettings(const std::string& FileName){
				InputLayer.SaveSettings(FileName + "-InputLayerSet.csv");
				HiddenLayer.SaveSettings(FileName + "-HiddenLayerSet.csv");
				OutputLayer.SaveSettings(FileName + "-OutputLayerSet.csv");
			}
			
			//! @brief CSVファイルから各レイヤのパーセプトロンの設定値を読み込む関数
			//! @param[in]	FileName	CSVファイル名(拡張子なし)
			void LoadSettings(const std::string& FileName){
				InputLayer.LoadSettings(FileName + "-InputLayerSet.csv");
				HiddenLayer.LoadSettings(FileName + "-HiddenLayerSet.csv");
				OutputLayer.LoadSettings(FileName + "-OutputLayerSet.csv");
			}
			
		private:
			MultiLayerPerceptron3(const MultiLayerPerceptron3&) = delete;					//!< コピーコンストラクタ使用禁止
			const MultiLayerPerceptron3& operator=(const MultiLayerPerceptron3&) = delete;	//!< 代入演算子使用禁止
			
			// ニューラルネットワークの基本設定
			static constexpr SlpDropout InDrop = SlpDropout::ENABLE;		//!< 入力層をドロップアウトするかどうか
			static constexpr SlpDropout HidDrop = SlpDropout::ENABLE;		//!< 隠れ層をドロップアウトするかどうか
			static constexpr SlpDropout InDropEst  = SlpDropout::ESTIMATE;	//!< 入力層でドロップアウト付き推定計算をするかどうか
			static constexpr SlpDropout HidDropEst = SlpDropout::ESTIMATE;	//!< 隠れ層でドロップアウト付き推定計算をするかどうか
			
			// 3層ニューラルネットワーク
			SingleLayerPerceptron<InSize, InSize, MinbatSize, InFunc> InputLayer;		//!< 入力層
			SingleLayerPerceptron<InSize, HidSize, MinbatSize, HidFunc> HiddenLayer;	//!< 隠れ層
			SingleLayerPerceptron<HidSize, OutSize, MinbatSize, OutFunc> OutputLayer;	//!< 出力層
			
			// 誤差逆伝播用の変数
			Matrix<MinbatSize, InSize> X;			//!< 順伝播入力行列(ミニバッチ)
			Matrix<MinbatSize, InSize> ZIn;			//!< 入力層の出力行列
			Matrix<MinbatSize, HidSize> ZHid;		//!< 内部層の出力行列
			Matrix<MinbatSize, OutSize> Y;			//!< 順伝播出力行列
			Matrix<MinbatSize, OutSize> D;			//!< 目標値行列
			Matrix<MinbatSize, HidSize> WDeltaOut;	//!< 出力層の重み誤差行列
			Matrix<MinbatSize, InSize> WDeltaHid;	//!< 内部層の重み誤差行列
			Matrix<MinbatSize, InSize> WDeltaIn;	//!< 入力層の重み誤差行列
			std::array<double, Epoch> EpochNumbers;	//!< エポック番号
			std::array<double, Epoch> TrainingError;//!< 訓練誤差
			std::array<double, Epoch> TestError;	//!< テスト誤差
			
			// テストデータ用の変数
			Matrix<MinbatSize, InSize> XTest;		//!< テスト用順伝播入力行列(ミニバッチ)
			Matrix<MinbatSize, InSize> ZInTest;		//!< 入力層の出力行列
			Matrix<MinbatSize, HidSize> ZHidTest;	//!< 内部層の出力行列
			Matrix<MinbatSize, OutSize> YTest;		//!< 順伝播出力行列
			Matrix<MinbatSize, OutSize> DTest;		//!< 目標値行列
			
			// 推定計算用の変数
			Matrix<1, InSize> zin;		//!< 入力層の出力ベクトル
			Matrix<1, HidSize> zhid;	//!< 内部層の出力ベクトル
	};

}

#endif

