//! @file TwoInertiaSimulator.cc
//! @brief 2慣性共振系シミュレータV2
//!
//! 2慣性共振系を模擬して，負荷側速度，ねじれ角，モータ速度を計算して出力する。
//!
//! @date 2020/04/03
//! @author Yokokura, Yuki
//
// Copyright (C) 2011-2020 Yuki YOKOKURA
// This program is free software;
// you can redistribute it and/or modify it under the terms of the FreeBSD License.
// For details, see the License.txt file.

#include <cassert>
#include <tuple>
#include "TwoInertiaParamDef.hh"
#include "TwoInertiaSimulator.hh"

// ARCS組込み用マクロ
#ifdef ARCS_IN
	// ARCSに組み込まれる場合
	#include "ARCSassert.hh"
	#include "ARCSeventlog.hh"
#else
	// ARCSに組み込まれない場合
	#define arcs_assert(a) (assert(a))
	#define PassedLog()
	#define EventLog(a)
	#define EventLogVar(a)
#endif

using namespace ARCS;

//! @brief コンストラクタ
//! @param[in]	Params	2慣性共振系パラメータ構造体
//! @param[in]	SmplTime	サンプリング周期 [s]
TwoInertiaSimulator::TwoInertiaSimulator(const struct TwoInertiaParamDef Params, const double SmplTime)
	: Kt(Params.Kt), Jm(Params.Jm), Jl(Params.Jl), Ks(Params.Ks),
	  Dm(Params.Dm), Dl(Params.Dl), Rg(Params.Rg), Ts(SmplTime), PlantSys()
{
	// 連続系A行列の設定
	const Matrix<3,3> A = {
		-Dm/Jm,      0, -Ks/(Rg*Jm),
		     0, -Dl/Jl,       Ks/Jl,
		1.0/Rg,     -1,           0
	};
	
	// 連続系B行列の設定
	const Matrix<2,3> B = {
		Kt/Jm,       0,
		    0, -1.0/Jl,
		    0,       0
	};
	
	const auto C = Matrix<3,3>::eye();		// C行列の設定
	PlantSys.SetContinuous(A, B, C, Ts);	// 状態空間モデルの設定＆離散化
	PassedLog();
}

//! @brief ムーブコンストラクタ
//! @param[in]	r	右辺値
TwoInertiaSimulator::TwoInertiaSimulator(TwoInertiaSimulator&& r)
	: Kt(r.Kt), Jm(r.Jm), Jl(r.Jl), Ks(r.Ks),
	  Dm(r.Dm), Dl(r.Dl), Rg(r.Rg), Ts(r.Ts), PlantSys(std::move(r.PlantSys))
{
	
}

//! @brief デストラクタ
TwoInertiaSimulator::~TwoInertiaSimulator(){
	PassedLog();
}

//! @brief 負荷側速度＆ねじれ角＆モータ側＆を計算する関数
//! @param[in]	Current	q軸電流 [A]
//! @param[in]	LoadTorque	負荷トルク [Nm]
//! @param[out]	LoadSpeed	負荷側速度 [rad/s]
//! @param[out]	TorsionAngle	ねじれ角 [rad]
//! @param[out]	MotorSpeed	モータ側速度 [rad/s]
void TwoInertiaSimulator::GetResponses(double Current, double LoadTorque, double& LoadSpeed, double& TorsionAngle, double& MotorSpeed){
	// 入力ベクトルの設定
	const Matrix<1,2> u = {
		Current,
		LoadTorque
	};
	
	Matrix<1,3> y = PlantSys.GetResponses(u);	// 状態空間モデルから出力を計算
	
	// 出力ベクトルから抽出
	MotorSpeed = y[1];
	LoadSpeed = y[2];
	TorsionAngle = y[3];
}

//! @brief 負荷側速度＆ねじれ角＆モータ側＆を計算する関数(タプル版)
//! @param[in]	Current	q軸電流 [A]
//! @param[in]	LoadTorque	負荷トルク [Nm]
//! @return	タプル( 負荷側速度 [rad/s], ねじれ角 [rad], モータ側速度 [rad/s] )
std::tuple<double,double,double> TwoInertiaSimulator::GetResponses(double Current, double LoadTorque){
	double MotorSpeed, LoadSpeed, TorsionAngle;
	GetResponses(Current, LoadTorque, LoadSpeed, TorsionAngle, MotorSpeed);
	return std::forward_as_tuple(LoadSpeed, TorsionAngle, MotorSpeed);
}

//! @brief 負荷側慣性を設定する関数
//! @param[in]	Inertia	負荷側慣性 [kgm^2]
void TwoInertiaSimulator::SetLoadInertia(double Inertia){
	Jl = Inertia;
}

//! @brief ばね定数を設定する関数
//! @param[in]	Stiffness ばね定数 [Nm/rad]
void TwoInertiaSimulator::SetSpring(double Stiffness){
	Ks = Stiffness;
}

//! @brief シミュレータをリセットする関数
void TwoInertiaSimulator::Reset(void){
	PlantSys.ClearStateVector();
}

