//! @file TwoInertiaSimulator.hh
//! @brief 2慣性共振系シミュレータV2
//!
//! 2慣性共振系を模擬して，負荷側速度，ねじれ角，モータ速度を計算して出力する。
//!
//! @date 2020/04/03
//! @author Yokokura, Yuki
//
// Copyright (C) 2011-2020 Yuki YOKOKURA
// This program is free software;
// you can redistribute it and/or modify it under the terms of the FreeBSD License.
// For details, see the License.txt file.

#ifndef TWOINERTIASIMULATOR
#define TWOINERTIASIMULATOR

#include "Matrix.hh"
#include "StateSpaceSystem.hh"

namespace ARCS {	// ARCS名前空間
	//! @brief 2慣性共振系シミュレータV2
	class TwoInertiaSimulator {
		public:
			TwoInertiaSimulator(const struct TwoInertiaParamDef Params, const double SmplTime);	//!< コンストラクタ
			TwoInertiaSimulator(TwoInertiaSimulator&& r);	//!< ムーブコンストラクタ
			~TwoInertiaSimulator();							//!< デストラクタ
			void GetResponses(double Current, double LoadTorque, double& LoadSpeed, double& TorsionAngle, double& MotorSpeed);	//!< 負荷側速度＆ねじれ角＆モータ側＆を計算する関数
			std::tuple<double,double,double> GetResponses(double Current, double LoadTorque);	//!< 負荷側速度＆ねじれ角＆モータ側＆を計算する関数(タプル版)
			void SetLoadInertia(double Inertia);	//!< 負荷側慣性を設定する関数
			void SetSpring(double Stiffness);		//!< ばね定数を設定する関数
			void Reset(void);						//!< シミュレータをリセットする関数
			
		private:
			TwoInertiaSimulator(const TwoInertiaSimulator&) = delete;					//!< コピーコンストラクタ使用禁止
			const TwoInertiaSimulator& operator=(const TwoInertiaSimulator&) = delete;	//!< 代入演算子使用禁止
			
			double Kt;	//!< [Nm/A]	トルク定数
			double Jm;	//!< [kgm^2]	モータ側慣性モーメント
			double Jl;	//!< [kgm^2]	負荷側慣性モーメント
			double Ks;	//!< [Nm/rad]	2慣性間の剛性
			double Dm;	//!< [Nm/(rad/s)]	モータ側粘性
			double Dl;	//!< [Nm/(rad/s)]	負荷側粘性
			double Rg;	//!< [-]	減速比
			double Ts;	//!< [s]	サンプリング時間
			StateSpaceSystem<3,2,3> PlantSys;	//!< 2慣性系システム
	};
}

#endif



